;;; -*- Mode:Lisp; Package:KEE; Base:10; Syntax: Common-lisp -*-
;;; ex: set lisp ai sm:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Copyright, 1990. The Regents of the University of California.
;;; This software was produced under a U. S. Government contract
;;; (W-7405-ENG-36) by Los Alamos National Laboratory, which is
;;; operated by the University of California for the U. S. Department
;;; of Energy.  The U. S. Government is licensed to use, reproduce,
;;; and distribute this software. Permission is granted to the
;;; public to copy and use this software without charge, provided
;;; that this Notice and any statement of authorship are reproduced
;;; on all copies.  Neither the Government nor the University makes
;;; any warranty, express or implied, or assumes any liability or
;;; responsibility for the use of this software.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
(in-package 'kee)

(goto.kb 'clos-on-kee)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; defclass is defined in full here. However, there is yet hair to be grown
;;; on standard-object. (In particular, the generic function initialize-instance
;;; will be added later.)
;;;
;;; The only parts of the metaobject protocol used here are the superclass
;;; of all user-defined classes, "standard-object", from which comes the
;;; generic function initialize-instance. This version of initialize-instance,
;;; while called by make-instance as per the spec, does not use any of the
;;; lower-level generic functions in the CLOS spec. It can, however, have before
;;; and after methods added to customize its behavior.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;
;;; We must manually create standard-object before any defclass is performed.
;;; This is the default superclass of any classes with no listed superclasses.
;;; (e.g. (defclass foo () ...)
;;; In other words, every class is a subclass of standard-object.
;;; We cannot create a "root" class until this "root of all roots" exists.
;;; If we had defclass at this early stage, this would be (almost)
;;;
;;; (defclass standard-object ()
;;;   ()
;;;   (:documentation "The superclass of all standard CLOS objects."))
;;;

(eval-when (compile load eval)
  (create.unit
	 'standard-object
	 'clos-on-kee
	 ()                   ; no KEE class parents
	 ()                   ; no KEE member parents
	 "The superclass of all standard CLOS objects."
	 ()                   ; no member slots
	 ()                   ; no own slots
	 ))


;;;
;;; We must manually create the class "standard-class" that is the default metaclass
;;; of things created here. This really should be
;;;
;;; (defclass standard-class (standard-object)
;;;   ((class-initarg-map
;;;     :documentation "an association list of initarg-name slot-name pairs."
;;;     :accessor class-initarg-map
;;;     :initform ())
;;;    (class-combined-initarg-map
;;;     :documentation "The initarg map inherited from all superclasses."
;;;     :accessor class-combined-initarg-map
;;;     :initform ())
;;;    (class-direct-initargs
;;;     :documentation "The default initargs from the defclass for this class."
;;;     :accessor class-direct-initargs
;;;     :initform ())
;;;    (class-combined-initargs
;;;     :documentation "All default initargs for this class."
;;;     :accessor class-combined-initargs
;;;     :initform ()))
;;;   (:documentation "The class of all class-instances created by defclass."))
;;;
;;; but we don't have defclass until at least one metaclass exists...
;;;
;;; standard-class is the parent class of any created classes. e.g. any user
;;; defined objects are subclasses of standard-object, but are instances of
;;; their metaclass (standard-class by default).
;;;
;;; After defmethod is defined, make-instance and allocate-instance can be
;;; attached, and defclass can be used to create new metaclasses.
;;;

(eval-when (compile load eval)
  (create.unit
	 'standard-class
	 'clos-on-kee
	 '(standard-object)                  ; subclass of standard-object
	 ()                                  ; no KEE member parents
	 "The class of all class-instances created by defclass."
	 (list
	  (list 'class-initarg-map
		(list nil)
		'override.values
		'nil
		nil
		'((comment
		   "an association list of initarg-name slot-name pairs.")))
	  (list 'class-combined-initarg-map
		(list nil)
		'override.values
		'nil
		nil
		'((comment
		   "The initarg map inherited from all superclasses.")))
	  (list 'class-direct-initargs
		(list nil)
		'override.values
		'nil
		nil
		'((comment
		   "The default initargs from the defclass for this class.")))
	  (list 'class-combined-initargs
		(list nil)
		'override.values
		'nil
		nil
		'((comment
		   "All default initargs for this class."))))
	 ())
  ;;
  ;; Now define the slot readers and setf methods as if the :accessor forms
  ;; had been used.
  ;;
  (defun class-direct-initargs (instance)
    (get.value instance
	'class-direct-initargs
	'own))
  (defun class-direct-initargs-setf-method (instance value)
    (put.value instance
	'class-direct-initargs
	value
	'own)
    value)
  (defsetf class-direct-initargs class-direct-initargs-setf-method)
  (defun class-combined-initargs (instance)
    (get.value instance
	'class-combined-initargs
	'own))
  (defun class-combined-initargs-setf-method (instance value)
    (put.value instance
	'class-combined-initargs
	value
	'own)
    value)
  (defsetf class-combined-initargs class-combined-initargs-setf-method)
  (defun class-initarg-map (instance)
    (get.value instance
	'class-initarg-map
	'own))
  (defun class-initarg-map-setf-method (instance value)
    (put.value instance
	'class-initarg-map
	value
	'own)
    value)
  (defsetf class-initarg-map class-initarg-map-setf-method)
  (defun class-combined-initarg-map (instance)
    (get.value instance
	'class-combined-initarg-map
	'own))
  (defun class-combined-initarg-map-setf-method (instance value)
    (put.value instance
	'class-combined-initarg-map
	value
	'own)
    value)
  (defsetf class-combined-initarg-map class-combined-initarg-map-setf-method)
  )

;;;
;;; ... and that last bunch of KEE code is why defclass exists in clos-on-kee.
;;;

;;;
;;; Create-slot takes a KEE slot specification and performs a create.slot
;;; to make the slot look like its defclass slot description. Create-slot
;;; removes the old slot (if it exists) before re-creating it. This is as
;;; close as we get to a re-definition protocol.
;;;

(defun create-slot (kb new-class slot-type slot-description)
  (let ((name (first slot-description))
	(init (second slot-description))
	(inherit (third slot-description))
	(type (fourth slot-description))
	(default (fifth slot-description))
	(facets (sixth slot-description)))
    (if (slot.exists.p name new-class kb)
	(delete.slot new-class name () type))
    (create.slot (list new-class kb) name slot-type init inherit type default facets t)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro defclass (new-class
		    superclasses
		    slot-list
		    &rest options)
  "provide a CLOS-like wrapper for KEE unit creation."
  (labels
      (
       ;;
       ;; Several functions break and examine the &rest option list.
       ;; The general form is
       ;;    loop through the options looking for one whose car is
       ;;    the desired option. Return the associated value(s).
       ;;    return a default value if no option is found.
       ;;
       
       ;;
       ;; :documentation
       ;;
       (get-doc-option (options)
	 (dolist (opt options "")
	   (if (eql (car opt) :documentation)
	       (return (second opt)))))
       ;;
       ;; :metaclass
       ;;
       (get-metaclass (options)
	 (dolist (opt options 'standard-class)
	   (if (eql (car opt) :metaclass)
	       (return (second opt)))))
       ;;
       ;; :default-initargs
       ;; The only trick is to run down the list two at a time, building
       ;; closures to save the lexical environment of this defclass.
       ;; fix-initargs is used for this.
       ;;
       (fix-initargs (initargs)
	 (if initargs
	     (cons
	      `(list ',(first initargs)
		#'(lambda () ,(second initargs)))
	      (fix-initargs (cddr initargs)))))
       (get-default-initargs (options)
	 (dolist (opt options ())
	   (if (eql (car opt) :default-initargs)
	       (return `(list ,@(fix-initargs (cdr opt)))))))
       ;;
       ;; parse-slot-options is the top routine called for each slot
       ;; description. It just determines if the slot description is a
       ;; single slot name "foo" or a slot specification such as
       ;; (foo :type this :initform ...) and calls build with
       ;; a proper common form
       ;;
       (parse-slot-options (type slotspec)
	 (if (atom slotspec)
	     (build slotspec type nil)
	     (build (car slotspec) type (cdr slotspec))))
       ;;
       ;; build does the work of parsing the slot options and returning
       ;; the proper form for each of the three places that it is called
       ;; from. The desired-type parameter is one of :instance
       ;; and :defun. This controls the two different sets of things
       ;; constructed.
       ;;
       ;; When :instance is performed, the slot options are parsed and
       ;; KEE slot specs are generated for each member slot.
       ;;
       ;; When :defun is performed, the slot descriptions are
       ;; parsed, and those that contain :reader :writer or :accessor
       ;; keys cause proper defun's to be generated for the desired access
       ;; functions and setfs.
       ;;
       (build (name desired-type option-list)
	 (let ((readers nil)
	       (writers nil)
	       (accessors nil)
	       (initargs nil)
	       (initform nil)
	       (type nil)
	       (doc-string nil))
	   (do ((options option-list (cddr options)))
	       ((null options))
	     (let ((option (car options))
		   (value (cadr options)))
	       (ecase option
		 (:reader (push value readers))
		 (:writer (push value writers))
		 (:accessor (push value readers)
			    (let ((n (gentemp)))
			      (push n writers)
			      (push (list value n) accessors)))
		 (:allocation (error ":allocation not supported in defclass"))
		 (:initarg (push (list value name) initargs))
		 (:initform (setf initform value))
		 (:type (setf type value))
		 (:documentation (setf doc-string value)))))
	   (ecase desired-type
	     (:class
	      ;;
	      ;; :class is used to extract each slot's initargs in order
	      ;; to construct the map of initarg-names to slot-names for this
	      ;; class. The value is just the list of initargs for this slot.
	      ;;
	      initargs)
	     (:instance
	      ;;
	      ;; :instance builds a KEE slot specification from our parts.
	      ;; This will be used either in a create.unit (if the class
	      ;; doesn't already exist) or in a create.slot within the
	      ;; create-slot procedure defined above.
	      ;;
	      `(list
		',name
		(list ,initform)
		'override.values
		',type
		nil
		'((comment ,doc-string))))
	     (:defun
		 ;;
		 ;; This pass over the slot descriptions generates defuns
		 ;; and defsetfs associated with :accessor, :reader, and
		 ;; :writer.
		 ;;
		 `(,@(mapcar #'(lambda (read-fnc)
				 `(defun ,read-fnc (instance)
				   (get.value instance ',name 'own)))
		      readers)
		   ,@(mapcar #'(lambda (write-fnc)
				 `(defun ,write-fnc (instance value)
				   (put.value instance ',name value 'own)
				   value))
		      writers)
		   ,@(mapcar #'(lambda (access-fnc-spec)
				 `(defsetf
				   ,(first access-fnc-spec)
				   ,(second access-fnc-spec)))
		      accessors)))))))
    ;;
    ;; main body
    ;;
    (let ((documentation (get-doc-option options))
	  (metaclass (get-metaclass options))
	  (member-slots (mapcar #'(lambda (slot-spec)
				    (parse-slot-options :instance slot-spec))
				slot-list))
	  (own-slots `(list
		       (list
			'class-direct-initargs
			,(let ((initargs (get-default-initargs options)))
			   (if initargs
			       `(list ,initargs)
			       ()))
			'override.values
			nil
			nil
			nil)
		       (list
			'class-initarg-map
			,(let ((map (mapcan #'(lambda (slot-spec)
						(parse-slot-options :class slot-spec))
					    slot-list)))
			   (if map
			       `'(,map)
			       ()))
			'override.values
			nil
			nil
			nil)))
	  (defuns (mapcan #'(lambda (slot-spec)
			      (parse-slot-options :defun slot-spec))
			  slot-list)))
      `(eval-when (compile load eval)
	(let ((new (gensym "TEMP-CLASS-")))
	  (create.unit
		 new
		 (kb)
		 ',(if (null superclasses)
		       (list '(standard-object clos-on-kee))
		       superclasses)
		 ;;
		 ;; Note that the following set of member parents enforces an
		 ;; important design decision:
		 ;; All metaclasses are units in the clos-on-kee kb.
		 ;;
		 '((,metaclass clos-on-kee))
		 ,(if documentation documentation "")
		 ,(if member-slots (cons 'list member-slots) nil)
		 ,own-slots)
	  (if (unit.exists.p ',new-class)
	      (delete.unit ',new-class new t new))
	  (rename.unit new ',new-class)
	  (unitmsg ',new-class 'initialize-instance)
	  ,@defuns
	  ',new-class)))))

;;;
;;; 'm' is for my debugging purposes only. Its purpose is to allow me to
;;; see how the first level of macro expansion looks (hence macroexpand-1
;;; rather than macroexpand).
;;;
;;; For example giving the lisp listener
;;;
;;;(m '(defclass foo (superclass1 superclass2)
;;;       (slot1
;;;	(slot2
;;;	 :documentation "A Second slot."
;;;	 :accessor get-slot2
;;;	 :type short-float
;;;	 :initform 23.0))
;;;       (:documentation "An foo-lish example class definition.")
;;;       (:metaclass my-metaclass)))
;;;
;;; Gives the following output:
;;;
;;;(EVAL-WHEN (COMPILE LOAD EVAL)
;;;           (LET ((NEW (GENSYM "TEMP-CLASS-")))
;;;             (CREATE.UNIT NEW
;;;                          (KB)
;;;                          '(SUPERCLASS1 SUPERCLASS2)
;;;                          '((MY-METACLASS CLOS-ON-KEE))
;;;                          "An foo-lish example class definition."
;;;                          (LIST
;;;                            (LIST 'SLOT1
;;;                                  (LIST NIL)
;;;                                  'OVERRIDE.VALUES
;;;                                  'NIL
;;;                                  NIL
;;;                                  '((COMMENT NIL)))
;;;                            (LIST 'SLOT2
;;;                                  (LIST 23.0)
;;;                                  'OVERRIDE.VALUES
;;;                                  'SHORT-FLOAT
;;;                                  NIL
;;;                                  '((COMMENT "A Second slot."))))
;;;                          NIL)
;;;             (IF (UNIT.EXISTS.P 'FOO)
;;;                 (DELETE.UNIT 'FOO NEW T NEW))
;;;             (RENAME.UNIT NEW 'FOO)
;;;             (UNITMSG 'FOO 'INITIALIZE-INSTANCE)
;;;             (DEFUN GET-SLOT2
;;;                    (INSTANCE)
;;;                    (GET.VALUE INSTANCE 'SLOT2 'OWN))
;;;             (DEFUN T2
;;;                    (INSTANCE VALUE)
;;;                    (PUT.VALUE INSTANCE 'SLOT2 VALUE 'OWN)
;;;                    VALUE)
;;;             (DEFSETF GET-SLOT2 T2)
;;;             'FOO))
;;;
;;;

(defun m (body) (pprint (macroexpand-1 body)))
