;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Copyright, 1990. The Regents of the University of California.
;;; This software was produced under a U. S. Government contract
;;; (W-7405-ENG-36) by Los Alamos National Laboratory, which is
;;; operated by the University of California for the U. S. Department
;;; of Energy.  The U. S. Government is licensed to use, reproduce,
;;; and distribute this software. Permission is granted to the
;;; public to copy and use this software without charge, provided
;;; that this Notice and any statement of authorship are reproduced
;;; on all copies.  Neither the Government nor the University makes
;;; any warranty, express or implied, or assumes any liability or
;;; responsibility for the use of this software.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; After the classes standard-class and standard-object are defined
;;; and defclass and defmethod macros are in place, the standard classes
;;; and objects can be fleshed out.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package 'kee)

;;;
;;; initialization of any instance
;;;

(eval-when (compile load eval)
  (defgeneric initialize-instance (self &rest initargs)))

(eval-when (compile load eval)
  (defmethod initialize-instance ((self standard-object) &rest initargs)
    "The default method for instance initialization."
    (let ((slot-map (class-combined-initarg-map (class-of self)))
	  (class (class-name (class-of self))))
      (labels
	  ((do-initargs (initargs)
	     (when initargs
	       (do-initargs (cddr initargs))
	       (let ((slot (second (assoc (car initargs) slot-map))))
		 (if (null slot)
		     (error "An initarg ~a intended to initialize an instance of~%~
                             class ~a has no slot associated with it. This means~%~
                             that the defclass form of class ~a, or some superclass~%~
                             of class ~a needs to mention initarg ~a in the :initarg~%~
                             option of some slot definition.~%"
			    (car initargs)
			    class
			    class
			    class
			    (car initargs))
		     (setf (slot-value self slot) (cadr initargs)))))))
	(do-initargs initargs)
    ))))

;;;
;;; allocate and make instances
;;;

(eval-when (compile load eval)
  (defgeneric allocate-instance (self &rest initargs)))

(eval-when (compile load eval)
  (defgeneric make-instance (self &rest initargs)))

(eval-when (compile load eval)
  (defmethod allocate-instance ((self standard-class) &rest initargs)
    "A simple CLOS compatable instance creator."
    (create.unit (gentemp (format nil "~a-" (unit.name self))) (kb) () (list self))))

(defun default-initargs (class initargs)
  "Build the defaulted-initargs list actually used for initialization."
  (labels
      (;;
       ;; initarg-exists is used to see if some default initarg has already
       ;; been added to the actual initarg list.
       ;;
       ;; initarg is the name of some initarg
       ;; initarg-list is a pairlist (name value name value ...)
       ;;
       (initarg-exists (initarg initarg-list)
	 (cond ((null initarg-list) nil)
	       ((eql initarg (car initarg-list)) t)
	       (t (initarg-exists initarg (cddr initarg-list)))))
       ;;
       ;; add-initarg is used to turn one of the combined initargs from the
       ;; class definition (initarg-name function-computing-value) into
       ;; an added "name value" pair in the actual initarg list.
       ;;
       (build-initargs (combined-initargs initargs)
	 (if (null combined-initargs)
	     initargs
	     (let* ((initarg-spec (car combined-initargs))
		    (name (first initarg-spec))
		    (function (second initarg-spec)))
	       (if (initarg-exists name initargs)
		   (build-initargs (cdr combined-initargs)
				   initargs)
		   (nconc (build-initargs (cdr combined-initargs)
					  initargs)
			  (list name (funcall function))))))))
					       
    (build-initargs (class-combined-initargs class)
		    (copy-tree initargs))))

(eval-when (compile load eval)
  (defmethod make-instance ((self standard-class) &rest initargs)
    "A simple CLOS compatable instance creator."
    (let ((actual-initargs (default-initargs self initargs)))
      (let ((instance (allocate-instance self actual-initargs)))
	(apply #'initialize-instance instance actual-initargs)
	instance))))

;;;
;;; Although a skeletal standard-class is defined directly with KEE as a
;;; part of the bootstrap process, a bit of the definition remains.
;;; In particular, we need an initialize-instance method on standard-class
;;; that cleans up the combined initarg information for each class whose
;;; metaclass is standard-class. Note that this combination can be performed
;;; by initialize-instance due to two things:
;;; 1. CLOS requires that all superclasses already exist when defclass is
;;;    performed. Hence we can get all the parent classes' default initargs and
;;;    slot-initarg name mapping information.
;;; 2. This package doesn't handle any class re-definition. We can get the
;;;    inherited information once and for all and forget about it.
;;;
;;; Also note that the (... &rest initargs) in the defmethod below is a little
;;; confusing. The parameter initargs are any initargs used by make-instance
;;; when creating the class (e.g. instance of standard-class). These are ignored
;;; here. The name initargs will be used through the rest of the method as a
;;; referent to the initarg definitions and the default-initargs definitions
;;; processed by the defclass for this instance of standard-class.
;;;

(defmethod initialize-instance :after ((self standard-class) &rest initargs)
  "Flesh out the initarg-related slots of a class."
  ;;
  ;; The slots class-direct-initargs and class-initarg-map contain association
  ;; lists pertinent to this class object. The combined forms are a merge of
  ;; all such from its superclasses.
  ;; The apply-append is used rather than a mapcan because we don't want to
  ;; destructively modify the superclass's slots.
  ;;
  (let ((superclasses (closure self #'class-direct-superclasses)))
    (setf (class-combined-initargs self)
	  (apply #'append (mapcar #'class-direct-initargs
				  superclasses)))
    (setf (class-combined-initarg-map self)
	  (apply #'append (mapcar #'class-initarg-map
				  superclasses)))
    ))
